@@ -21,13 +21,13 @@ from jupyter_client import KernelManager, AsyncKernelManager
 from subprocess import PIPE
 
 from ..manager import start_new_kernel, start_new_async_kernel
-from .utils import test_env, skip_win32, AsyncKernelManagerSubclass, AsyncKernelManagerWithCleanup
+from ..manager import _ShutdownStatus
+from .utils import test_env, SyncKernelManagerSubclass, AsyncKernelManagerSubclass, AsyncKernelManagerWithCleanup
 
 pjoin = os.path.join
 
 TIMEOUT = 30
 
-
 @pytest.fixture(autouse=True)
 def env():
     env_patch = test_env()
@@ -52,18 +52,42 @@ def config(transport):
     return c
 
 
+def _install_kernel(name="signaltest", extra_env=None):
+    if extra_env is None:
+        extra_env = dict()
+    kernel_dir = pjoin(paths.jupyter_data_dir(), "kernels", name)
+    os.makedirs(kernel_dir)
+    with open(pjoin(kernel_dir, "kernel.json"), "w") as f:
+        f.write(
+            json.dumps(
+                {
+                    "argv": [
+                        sys.executable,
+                        "-m",
+                        "jupyter_client.tests.signalkernel",
+                        "-f",
+                        "{connection_file}",
+                    ],
+                    "display_name": "Signal Test Kernel",
+                    "env": {"TEST_VARS": "${TEST_VARS}:test_var_2", **extra_env},
+                }
+            )
+        )
+
+
 @pytest.fixture
 def install_kernel():
-    kernel_dir = pjoin(paths.jupyter_data_dir(), 'kernels', 'signaltest')
-    os.makedirs(kernel_dir)
-    with open(pjoin(kernel_dir, 'kernel.json'), 'w') as f:
-        f.write(json.dumps({
-            'argv': [sys.executable,
-                     '-m', 'jupyter_client.tests.signalkernel',
-                     '-f', '{connection_file}'],
-            'display_name': "Signal Test Kernel",
-            'env': {'TEST_VARS': '${TEST_VARS}:test_var_2'},
-        }))
+    return _install_kernel()
+
+
+def install_kernel_dont_shutdown():
+    _install_kernel("signaltest-no-shutdown", {"NO_SHUTDOWN_REPLY": "1"})
+
+
+def install_kernel_dont_terminate():
+    return _install_kernel(
+        "signaltest-no-terminate", {"NO_SHUTDOWN_REPLY": "1", "NO_SIGTERM_REPLY": "1"}
+    )
 
 
 @pytest.fixture
@@ -75,32 +99,16 @@ def start_kernel():
     assert km.context.closed
 
 
-@pytest.fixture
-def start_kernel_w_env():
-    kernel_cmd = [sys.executable,
-                  '-m', 'jupyter_client.tests.signalkernel',
-                  '-f', '{connection_file}']
-    extra_env = {'TEST_VARS': '${TEST_VARS}:test_var_2'}
-
-    km = KernelManager(kernel_name='signaltest')
-    km.kernel_cmd = kernel_cmd
-    km.extra_env = extra_env
-    km.start_kernel()
-    kc = km.client()
-    kc.start_channels()
-
-    kc.wait_for_ready(timeout=60)
-
-    yield km, kc
-    kc.stop_channels()
-    km.shutdown_kernel()
-
-
 @pytest.fixture
 def km(config):
     km = KernelManager(config=config)
     return km
 
+@pytest.fixture
+def km_subclass(config):
+    km = SyncKernelManagerSubclass(config=config)
+    return km
+
 
 @pytest.fixture
 def zmq_context():
@@ -116,6 +124,12 @@ def async_km(request, config):
     return km
 
 
+@pytest.fixture
+def async_km_subclass(config):
+    km = AsyncKernelManagerSubclass(config=config)
+    return km
+
+
 @pytest.fixture
 @async_generator  # This is only necessary while Python 3.5 is support afterwhich both it and yield_() can be removed
 async def start_async_kernel():
@@ -126,6 +140,55 @@ async def start_async_kernel():
     assert km.context.closed
 
 
+class TestKernelManagerShutDownGracefully:
+    parameters = (
+        "name, install, expected",
+        [
+            ("signaltest", _install_kernel, _ShutdownStatus.ShutdownRequest),
+            (
+                "signaltest-no-shutdown",
+                install_kernel_dont_shutdown,
+                _ShutdownStatus.SigtermRequest,
+            ),
+            (
+                "signaltest-no-terminate",
+                install_kernel_dont_terminate,
+                _ShutdownStatus.SigkillRequest,
+            ),
+        ],
+    )
+
+    @pytest.mark.skipif(
+        sys.platform == "win32", reason="Windows doesn't support signals"
+    )
+    @pytest.mark.parametrize(*parameters)
+    def test_signal_kernel_subprocesses(self, name, install, expected):
+        install()
+        km, kc = start_new_kernel(kernel_name=name)
+        assert km._shutdown_status == _ShutdownStatus.Unset
+        assert km.is_alive()
+        # kc.execute("1")
+        kc.stop_channels()
+        km.shutdown_kernel()
+
+        assert km._shutdown_status == expected
+
+    @pytest.mark.skipif(
+        sys.platform == "win32", reason="Windows doesn't support signals"
+    )
+    @pytest.mark.parametrize(*parameters)
+    async def test_async_signal_kernel_subprocesses(self, name, install, expected):
+        install()
+        km, kc = await start_new_async_kernel(kernel_name=name)
+        assert km._shutdown_status == _ShutdownStatus.Unset
+        assert await km.is_alive()
+        # kc.execute("1")
+        kc.stop_channels()
+        await km.shutdown_kernel()
+
+        assert km._shutdown_status == expected
+
+
 class TestKernelManager:
 
     def test_lifecycle(self, km):
@@ -154,8 +217,11 @@ class TestKernelManager:
         km, kc = start_kernel
 
         def execute(cmd):
-            kc.execute(cmd)
-            reply = kc.get_shell_msg(TIMEOUT)
+            request_id = kc.execute(cmd)
+            while True:
+                reply = kc.get_shell_msg(TIMEOUT)
+                if reply['parent_header']['msg_id'] == request_id:
+                    break
             content = reply['content']
             assert content['status'] == 'ok'
             return content
@@ -193,8 +259,11 @@ class TestKernelManager:
 
     def _env_test_body(self, kc):
         def execute(cmd):
-            kc.execute(cmd)
-            reply = kc.get_shell_msg(TIMEOUT)
+            request_id = kc.execute(cmd)
+            while True:
+                reply = kc.get_shell_msg(TIMEOUT)
+                if reply['parent_header']['msg_id'] == request_id:
+                    break
             content = reply['content']
             assert content['status'] == 'ok'
             return content
@@ -210,13 +279,6 @@ class TestKernelManager:
         assert km.context.closed is False
         self._env_test_body(kc)
 
-    def test_templated_extra_env(self, install_kernel, start_kernel_w_env):
-        km, kc = start_kernel_w_env
-        assert km.is_alive()
-        assert kc.is_alive()
-        assert km.context.closed is False
-        self._env_test_body(kc)
-
     def test_cleanup_context(self, km):
         assert km.context is not None
         km.cleanup_resources(restart=False)
@@ -232,6 +294,45 @@ class TestKernelManager:
         assert km.context.closed is False
         assert zmq_context.closed is False
 
+    def test_subclass_callables(self, km_subclass):
+        km_subclass.reset_counts()
+        km_subclass.start_kernel(stdout=PIPE, stderr=PIPE)
+        assert km_subclass.call_count('start_kernel') == 1
+        assert km_subclass.call_count('_launch_kernel') == 1
+
+        is_alive = km_subclass.is_alive()
+        assert is_alive
+        km_subclass.reset_counts()
+        km_subclass.restart_kernel(now=True)
+        assert km_subclass.call_count('restart_kernel') == 1
+        assert km_subclass.call_count('shutdown_kernel') == 1
+        assert km_subclass.call_count('interrupt_kernel') == 1
+        assert km_subclass.call_count('_kill_kernel') == 1
+        assert km_subclass.call_count('cleanup_resources') == 1
+        assert km_subclass.call_count('start_kernel') == 1
+        assert km_subclass.call_count('_launch_kernel') == 1
+
+        is_alive = km_subclass.is_alive()
+        assert is_alive
+
+        km_subclass.reset_counts()
+        km_subclass.interrupt_kernel()
+        assert km_subclass.call_count('interrupt_kernel') == 1
+
+        assert isinstance(km_subclass, KernelManager)
+
+        km_subclass.reset_counts()
+        km_subclass.shutdown_kernel(now=False)
+        assert km_subclass.call_count('shutdown_kernel') == 1
+        assert km_subclass.call_count('interrupt_kernel') == 1
+        assert km_subclass.call_count('request_shutdown') == 1
+        assert km_subclass.call_count('finish_shutdown') == 1
+        assert km_subclass.call_count('cleanup_resources') == 1
+
+        is_alive = km_subclass.is_alive()
+        assert is_alive is False
+        assert km_subclass.context.closed
+
 
 class TestParallel:
 
@@ -302,8 +403,11 @@ class TestParallel:
         kc = self._prepare_kernel(km, stdout=PIPE, stderr=PIPE)
 
         def execute(cmd):
-            kc.execute(cmd)
-            reply = kc.get_shell_msg(TIMEOUT)
+            request_id = kc.execute(cmd)
+            while True:
+                reply = kc.get_shell_msg(TIMEOUT)
+                if reply['parent_header']['msg_id'] == request_id:
+                    break
             content = reply['content']
             assert content['status'] == 'ok'
             return content
@@ -372,8 +476,11 @@ class TestAsyncKernelManager:
         km, kc = start_async_kernel
 
         async def execute(cmd):
-            kc.execute(cmd)
-            reply = await kc.get_shell_msg(TIMEOUT)
+            request_id = kc.execute(cmd)
+            while True:
+                reply = await kc.get_shell_msg(TIMEOUT)
+                if reply['parent_header']['msg_id'] == request_id:
+                    break
             content = reply['content']
             assert content['status'] == 'ok'
             return content
@@ -388,10 +495,13 @@ class TestAsyncKernelManager:
         assert reply['user_expressions']['poll'] == [None] * N
 
         # start a job on the kernel to be interrupted
-        kc.execute('sleep')
+        request_id = kc.execute('sleep')
         await asyncio.sleep(1)  # ensure sleep message has been handled before we interrupt
         await km.interrupt_kernel()
-        reply = await kc.get_shell_msg(TIMEOUT)
+        while True:
+            reply = await kc.get_shell_msg(TIMEOUT)
+            if reply['parent_header']['msg_id'] == request_id:
+                break
         content = reply['content']
         assert content['status'] == 'ok'
         assert content['user_expressions']['interrupted'] is True
@@ -412,3 +522,42 @@ class TestAsyncKernelManager:
         assert is_alive
         is_alive = await kc.is_alive()
         assert is_alive
+
+    async def test_subclass_callables(self, async_km_subclass):
+        async_km_subclass.reset_counts()
+        await async_km_subclass.start_kernel(stdout=PIPE, stderr=PIPE)
+        assert async_km_subclass.call_count('start_kernel') == 1
+        assert async_km_subclass.call_count('_launch_kernel') == 1
+
+        is_alive = await async_km_subclass.is_alive()
+        assert is_alive
+        async_km_subclass.reset_counts()
+        await async_km_subclass.restart_kernel(now=True)
+        assert async_km_subclass.call_count('restart_kernel') == 1
+        assert async_km_subclass.call_count('shutdown_kernel') == 1
+        assert async_km_subclass.call_count('interrupt_kernel') == 1
+        assert async_km_subclass.call_count('_kill_kernel') == 1
+        assert async_km_subclass.call_count('cleanup_resources') == 1
+        assert async_km_subclass.call_count('start_kernel') == 1
+        assert async_km_subclass.call_count('_launch_kernel') == 1
+
+        is_alive = await async_km_subclass.is_alive()
+        assert is_alive
+
+        async_km_subclass.reset_counts()
+        await async_km_subclass.interrupt_kernel()
+        assert async_km_subclass.call_count('interrupt_kernel') == 1
+
+        assert isinstance(async_km_subclass, AsyncKernelManager)
+
+        async_km_subclass.reset_counts()
+        await async_km_subclass.shutdown_kernel(now=False)
+        assert async_km_subclass.call_count('shutdown_kernel') == 1
+        assert async_km_subclass.call_count('interrupt_kernel') == 1
+        assert async_km_subclass.call_count('request_shutdown') == 1
+        assert async_km_subclass.call_count('finish_shutdown') == 1
+        assert async_km_subclass.call_count('cleanup_resources') == 1
+
+        is_alive = await async_km_subclass.is_alive()
+        assert is_alive is False
+        assert async_km_subclass.context.closed
